--  2006 Peter Thiemann
{- | String-based implementation of transactional database access.
-}
module WASH.DB.Transaction
  (
    -- * Database connection
    D (), register,
    -- * Database tuple handles
    T (), findByPK, findByNV, find, getPrimaryKey,
    -- * Operations on tuples
    create, remove, 
    get, getMany, getAll, set, setMany, 
    -- * Transaction support
    withState, State (..),
    with, Control (..),
    TCGI ()
  )
where

import qualified Database.HDBC as DB
-- import qualified Database.TRANSDB.TraceQuery as DB

-- transaction support independent of CGI
import qualified Database.TRANSDB.Manager as DBM
import qualified Database.TRANSDB.SQL as SQL
import Database.TRANSDB.LowLevel
import Database.TRANSDB.Backend

import System.IO.Unsafe (unsafePerformIO)
import Data.IORef
import qualified List
import Maybe (fromJust, isNothing, isJust, maybeToList)
import Monad (when, mplus)

import WASH.CGI.TCGI

import qualified WASH.CGI.HTMLWrapper as H hiding (map,head)
import qualified WASH.CGI.BaseCombinators as B
import WASH.CGI.CGIMonad
import qualified WASH.CGI.TransactionUtil as TU
import WASH.CGI.TransactionUtil (State (..), Control (..))
import WASH.CGI.LogEntry

-- | Abstract type of database handles. Contains a hashvalue for the db.
type D = DBM.D

-- | Register a database for use in a transaction. Takes a connection specifier and
-- a function that creates the connection.
-- Returns a database handle. Just a wrapper for 'DBM.register'.
-- 
-- Example usage: @register \"dbname=exampledb\" DB.connectPostgreSQL@
register :: (CGIMonad cgi, Show cs) => cs -> (cs -> IO DB.Connection) -> cgi D
register connectionSpec connect =
  wrapCGI (\cgistate ->
    do hash <- DBM.registerIO connectionSpec connect
       case inparm cgistate of
	PAR_IGNORED : rest ->
	  return (hash, cgistate { inparm = rest })
	[] ->
	  let stid = stateID cgistate
	      newparm = PAR_IGNORED in
          return (hash, cgistate { outparm = newparm : outparm cgistate
				 , stateID = nextstid stid newparm })
	_ ->
	  B.reportError "registerDB: out of sync" H.empty cgistate)

-- |Create a fresh tuple in @table@ with new primary key and set its initial
-- contents. Returns a handle for the tuple.
-- 
-- Complication: the fresh tuple must not be visible outside this transaction
-- but we need to ensure that its unique primary key is reserved, anyway. Hence,
-- we just reserve it through the sequence first time round.
create :: D -> TableName -> [NamedValue] -> TCGI b T
create db table namedValues =
  wrapCGI (\cgistate ->
    let inp = inparm cgistate
    in
    case map fromPAR inp of
      [] -> do	-- first time here
        conn <- DBM.getCachedConnection db
	pk <- generatePrimaryKey conn table
	checkFieldsForExistence db table namedValues
	let tid = T db table pk 
	    newparm = mkPAR $ NEW_TUPLE tid (List.sort namedValues)
	return (tid,
		cgistate { outparm = newparm : outparm cgistate })
      Right (NEW_TUPLE tid namedValues): rest ->	-- second, third, etc times
	return (tid,
		cgistate { inparm = tail inp })
      _ ->
	B.reportError "DB.Transaction.new: out of sync" H.empty cgistate)

-- |Remove tuple identified by a primary key from DB. Only registers the action
-- in the log, but checks that the tuple exists.
-- Fails (with an exception) if the tuple does not exist.
remove :: T -> TCGI b ()
remove tid =
  checkTupleExists tid False >>= \ tupleExists ->
  wrapCGI (\cgistate ->
    let inp = inparm cgistate
	out = outparm cgistate
	stid = stateID cgistate
    in
    case map fromPAR inp of
      [] ->	-- first time
        if tupleExists 
	then 
	  let newparm = mkPAR $ REM_TUPLE tid in
	  return ((),
		  cgistate { outparm = newparm : out
			   , stateID = nextstid stid newparm })
	else
	  error ("DB.Transaction.remove: tuple " ++ show tid ++ " not present")
      Right (REM_TUPLE tid') : _ ->
	return ((),
		cgistate { inparm = tail inp })
      _ ->
	B.reportError "DB.Transaction.remove: out of sync" H.empty cgistate
    )


-- |Check existence of a tuple.
checkTupleExists :: T -> Bool -> TCGI b Bool
checkTupleExists tid doRead =
  wrapCGI (\cgistate ->
    let inp = inparm cgistate
	out = outparm cgistate
	stid = stateID cgistate
    in
    case map fromPAR inp of
      [] ->
        let log = filterPAR out 
	    tupleIDStatus = checkTupleIDStatus tid log
	in
	case tupleIDStatus of
	  Exists ->
	    let newparm = PAR_RESULT (show True) in
	    return (True, cgistate { outparm = newparm : out })
	  Removed ->
	    let newparm = PAR_RESULT (show False) in
	    return (False, cgistate { outparm = newparm : out })
	  Unknown ->
	    if doRead 
	    then
	      do nvs <- reallyReadValues tid
		 let pkColumn = primaryKeyColumn (t_tableName tid)
		 case lookup pkColumn nvs of
		   Just (Just _) ->
		     let newparm = mkPAR $ READ_TUPLE tid nvs [] [] in
		     return (True, cgistate { outparm = newparm : out })
		   _ ->
		     let newparm = PAR_RESULT (show False) in
		     return (False, cgistate { outparm = newparm : out })
	    else
	      do exists <- reallyCheckExists tid
		 let newparm = PAR_RESULT (show exists)
		 return (exists, cgistate { outparm = newparm : out })
      Right (READ_TUPLE {}) : _ ->
	return (True, cgistate { inparm = tail (inparm cgistate) })
      Left (PAR_RESULT bool) : _ ->
	return (read bool, cgistate { inparm = tail (inparm cgistate) })
      _ ->
	B.reportError "DB.Transaction.checkTupleExists: out of sync" H.empty cgistate
    )

data TupleIDExistence =
    Exists | Removed | Unknown

-- |does this tuple ID exist in the log?
checkTupleIDStatus :: T -> [LogEntry] -> TupleIDExistence
checkTupleIDStatus tid les =
  foldr g Unknown les
  where
    g :: LogEntry -> TupleIDExistence -> TupleIDExistence
    g le b =
      if le_tid le /= tid 
      then b
      else -- this log entry is for our tuple
      case le of
	REM_TUPLE {} ->
	  Removed
	_ ->
	  Exists

-- |Obtain a tuple through its primary key. Checks for its existence and
-- preloads the tuple in the cache.
findByPK :: D -> TableName -> PrimaryKey -> TCGI b (Maybe T)
findByPK db table pk =
  let tid = T db table pk in
  checkTupleExists tid True >>= \ tupleExists ->
  if tupleExists 
  then return (Just tid)
  else return Nothing
  
-- |Obtain tuples through a where-clause of the form
-- name_i=value_i. Preloads. Caution: may generate an unlimited number of log
-- entries. 
-- 
-- TODO: Should perhaps take a @LIMIT@ parameter
findByNV :: D -> TableName -> [NamedValue] -> String -> TCGI b [T]
findByNV db table nvs extraSQL =
  wrapCGI (\cgistate ->
    let inp = inparm cgistate
	out = outparm cgistate
	g mappedInp inp ret =
	  case mappedInp of
	    [] | null ret -> 
	      do allReads <- reallyReadTuples db table [] nvs extraSQL
		 let newparms = map mkPAR (END_TUPLE : allReads)
		 return (map le_tid allReads,
			 cgistate { outparm = newparms ++ out })
	    Right (le@READ_TUPLE {}) : rest ->
	      g rest (tail inp) (le_tid le : ret)
	    Right (END_TUPLE) : rest ->
	      return (ret,
		      cgistate { inparm = tail inp })
	    _ ->
	      B.reportError "DB.Transaction.findByNV: out of sync" H.empty cgistate
    in 
    g (map fromPAR inp) inp [])

-- |Take a database, a table, and an SQL where-clause. 
-- Return a list of tuple IDs for the results.
find :: D -> TableName -> WhereClause -> TCGI b [T]
find db table whereClause =
  do pks <- B.once (wrapCGI (wrapIO (findIO db table whereClause)))
     let makeT pk = T db table pk
     return $ map makeT pks

-- |Get the primary key from a tuple handle. Does not access the DB at all. Does
-- not affect the transactional behavior of the tuple (it does not count as a
-- read operation on the tuple).
getPrimaryKey :: T -> DBValue
getPrimaryKey tid =
  Just (t_primaryKey tid)

-- |Get field value from handle. Caches the whole tuple.
get :: T -> ColumnName -> TCGI b (DBValue)
get tid field =
  do [value] <- getMany tid [field]
     return value

-- |Get field values from handle. Caches the whole tuple.
getMany :: T -> [ColumnName] -> TCGI b [DBValue]
getMany tid colNames =
  do nvs <- getInternal tid (readValues' tid colNames)
     return (map snd nvs)

getInternal :: 
  T -> ([LogEntry] -> IO ([NamedValue], [LogEntry])) -> TCGI b [NamedValue]
getInternal tid readValues =
  wrapCGI (\ cgistate ->
    let inp = inparm cgistate in
    case map fromPAR inp of
      [] ->	-- first time
        do (values, newparmsLE) <- readValues (filterPAR $ outparm cgistate)
	   let newparms = map mkPAR newparmsLE
	       newparm = head newparms
	       stid = stateID cgistate
	   return (values, 
		   cgistate { outparm = newparms ++ outparm cgistate
			    , stateID = nextstid stid newparm })
      Right (READ_TUPLE {}) 
	: Right (GET_TUPLE tid'' namedValues'') : rest ->
	-- next times, redo initial read
	return (namedValues'',
		cgistate { inparm = tail (tail inp) })
      Right (GET_TUPLE tid' namedValues') : rest ->
	-- next times, redo subseqent read
	return (namedValues',
		cgistate { inparm = tail inp })
      _ ->
	B.reportError "DB.Transaction.get: out of sync" H.empty cgistate
  )

-- |Get whole tuple from handle.
getAll :: T -> TCGI b [NamedValue]
getAll tid =
  getInternal tid (readAllValues tid)

-- |Set field value through handle.
set :: T -> ColumnName -> DBValue -> TCGI b ()
set handle field val =
  -- trivial implementation
  setMany handle [(field, val)]

-- |Set many columns at once through handle
setMany :: T -> [NamedValue] -> TCGI b ()
setMany tid namedValues =
  wrapCGI (\cgistate ->
    -- log manipulation, only
    let inp = inparm cgistate in
    case map fromPAR inp of
      [] ->
        return ((), cgistate { outparm = mkPAR (SET_TUPLE tid namedValues) : outparm cgistate })
      Right (SET_TUPLE _ _) : rest ->
	return ((), cgistate { inparm = tail inp })
      _ ->
	B.reportError "DB.Transaction.get: out of sync" H.empty cgistate
  )

-- |Starts a transactional scope. Same interface as
-- 'WASH.CGI.Transaction.withState'. @rollback@ be buggy.
withState :: (Read state, Show state) =>
  state 			--  initial state
  -> (State 
        -> state 			--  current state
        -> (Maybe state -> CGI ())	--  rollback continuation
	-> (state -> CGI () -> CGI ())	--  invoke commit continuation
	-> CGI ())		
  -> CGI ()
withState =
  TU.withState commitFromPARAMETER TU.Initial

-- |Starts a transactional scope. Is called with a default result, a
-- continuation that takes a result and yields a @CGI@ action, and the function
-- implementing the body of the scope. Its 'Control'
-- parameter is a tuple of functions to control the transaction. If @ta ::
-- Control@, then @abandon ta result@ exits the transaction without committing
-- and invokes the continuation with @result@. @setFail ta result@ and @setCommit ta
-- result@ set the result value in case of a failure (or success) to commit the
-- transaction. If the body function runs to the end, it tries to commit the
-- transaction and invokes the respective continuation.
-- 
-- Flexible transaction. do the right thing depending on the monad in which we
-- are running.
with :: (CGIMonad cgi, Read result, Show result) =>
     result				-- default result
  -> (result -> cgi ())		-- result continuation
  -> (Control (TCGI result) result -> TCGI result ())
  -> cgi ()
with result onResult fun =
  do cfl <- chooser
              commitFromPARAMETER   -- unnested case: do the real thing
	      (const $ return True) -- nested case: always suceed, do nothing
     TU.with cfl result onResult fun

-- |Adapter function to 'commitFromLog'
commitFromPARAMETER :: [PARAMETER] -> IO Bool
commitFromPARAMETER =
  commitFromLog . filterPAR

-- |embeds a 'LogEntry' into 'PARAMETER'
mkPAR :: LogEntry -> PARAMETER
mkPAR le =
  PAR_DB (show le)

-- |extracts a 'LogEntry' from a 'PARAMETER'
fromPAR :: PARAMETER -> Either PARAMETER LogEntry
fromPAR (PAR_DB str) = 
  case reads str of
    (v, ""): _ ->
      Right v
    _ ->
      error ("DB.Transaction.fromPAR: unreadable " ++ show str)
fromPAR p =
  Left p

-- |filters 'LogEntry' records from a list of 'PARAMETER'
filterPAR :: [PARAMETER] -> [LogEntry]
filterPAR =
  concatMap (maybeToList . eitherToMaybe . fromPAR)

-- |Injects values from the @Either a@ monad into the @Maybe@ monad by ignoring
-- the @Left@ values.
eitherToMaybe :: Either a b -> Maybe b
eitherToMaybe (Left x) = Nothing
eitherToMaybe (Right x) = Just x
